#define PI 3.141592653589793238462643383279502884197

/* 係数 */
inline double funCF(int k,double val)
    {
      if(k==0){
	return val;
      }else{
	return 1.0;
      }
    }

/* Cosine基底関数 */
inline double getCos(int j,int u,int block)
{
  return cos((((2.0*((double)(j))+1.0)*(((double)(u))*PI))/
	      (2.0*((double)(block)))));
}

/* ブロック(部分画像)のDCT */
void DCTSub(Image *in,Image *out,Image *weight1,Image *weight2){
  
  int u,v,j,k;
  for(u=0;u<in->sy;u++)
    for(v=0;v<in->sx;v++){
      double total=0.0;
      for(j=0;j<in->sy;j++)
	for(k=0;k<in->sx;k++)
	  total += in->img[k][j]*weight1->img[(u*in->sx+v)][(j*in->sx+k)];
      out->img[u][v] = weight2->img[u][v]*total; 
    }
}

/* ブロック(部分画像)の逆DCT */
void InverseDCTSub(Image *in,Image *out,Image *weight1,double TwoSQRTSubXSubYIn){
  
  int u,v,j,k;
  for(j=0;j<in->sy;j++)
    for(k=0;k<in->sx;k++){
      double total=0.0;
      for(u=0;u<in->sy;u++)
	for(v=0;v<in->sx;v++)
	  total += in->img[v][u]*weight1->img[(u*in->sx+v)][(j*in->sx+k)];
      out->img[j][k] = TwoSQRTSubXSubYIn*total; 
    }
}


/* 画像のDCTを横subX縦subYのブロック(部分画像)毎に実効する関数 */
Image * DCT(Image *in,int subX,int subY){
  int nx,ny;
  int i,j,k,l,m;
  Image *subIn = new Image(subX,subY);
  Image *subOut = new Image(subX,subY);
  Image *weight1 = new Image(subX*subY,subX*subY);
  Image *weight2 = new Image(subX,subY);
  

  /* cos()やsqrt()等の数学関数の呼び出しを最小限に抑える->高速化のための重み配列作成 */
  double SQRT2In = 1.0/sqrt(2.0);
  double TwoSQRTSubXSubYIn = 2.0/sqrt((double)(subX*subY));
  int u,v;
  /* 係数 */
  for(u=0;u<subY;u++)
    for(v=0;v<subX;v++){
      weight2->img[u][v] = TwoSQRTSubXSubYIn*funCF(u,SQRT2In)*funCF(v,SQRT2In);
      
    }
  /* 基底 */
  for(u=0;u<subY;u++)
    for(v=0;v<subX;v++)
      for(j=0;j<subY;j++)
	for(k=0;k<subX;k++)
	  weight1->img[(u*subX+v)][(j*subX+k)]= (getCos(j,v,subY)*getCos(k,u,subX));
  
  
  /* 画像の縦横を8の倍数に正規化する(切り捨てか0や端の値を埋め込む、ここでは端の値を埋め込み)。*/
  
  nx = in->sx + (subX-(in->sx % subX));
  ny = in->sy + (subY-(in->sy % subY));
  
  Image *dummyIn = new Image(nx,ny);
  Image *dummyOut = new Image(nx,ny);
  
  /* 初期化 */
  for(i=0;i<ny;i++)
    for(j=0;j<nx;j++){
      dummyIn->img[i][j] = 0.0;
      dummyOut->img[i][j] = 0.0;
    }
  for(i=in->sy;i<ny;i++)
    for(j=0;j<in->sx;j++){
      dummyIn->img[i][j] = in->img[in->sy-1][j];
    }
  for(i=0;i<in->sy;i++)
    for(j=in->sx;j<nx;j++){
      dummyIn->img[i][j] = in->img[i][in->sx-1];
    }

    
  for(i=0;i<in->sy;i++)
    for(j=0;j<in->sx;j++)
      dummyIn->img[i][j] = in->img[i][j];
  
  /* mainの subX by subYブロックのloop */
  for(i=0;i<ny;i+=subY)
    for(j=0;j<nx;j+=subX){
      
      /* 画像をsubXxsubYの部分画像に分割する。*/
      for(l=0;l<subY;l++)
	for(m=0;m<subX;m++)
	  subIn->img[l][m] = dummyIn->img[(i+l)][(j+m)];
      
      
      /*  部分画像１つに対してDCTを行う。*/
      DCTSub(subIn,subOut,weight1,weight2);
      
      for(l=0;l<subY;l++)
	for(m=0;m<subX;m++){
	  /* ↓DCTされた部分画像の周波数を保存用の配列に統合する。*/
	  dummyOut->img[(i+l)][(j+m)] = subOut->img[l][m];
	}
    }
  
  
  
  delete weight1;
  delete weight2;
  delete subIn;
  delete subOut;
  delete dummyIn;
  
  return dummyOut;
  
}
/* 画像の逆DCTを横subX縦subYのブロック(部分画像)毎に実効する関数 */
void InverseDCT(Image *in,Image *out,int subX,int subY){
  int nx,ny;
  int i,j,k,l,m;
  Image *subIn = new Image(subX,subY);
  Image *subOut = new Image(subX,subY);
  Image *weight1 = new Image(subX*subY,subX*subY);
 
  

  /* cos()やsqrt()等の数学関数の呼び出しを最小限に抑える->高速化のための重み配列作成 */
  double SQRT2In = 1.0/sqrt(2.0);
  double TwoSQRTSubXSubYIn = 2.0/sqrt((double)(subX*subY));
  int u,v;
  
  
  /* 基底(Inverse) */
   for(j=0;j<subY;j++)
     for(k=0;k<subX;k++)
       for(u=0;u<subY;u++)
	 for(v=0;v<subX;v++)
	   weight1->img[(u*subX+v)][(j*subX+k)]= (funCF(u,SQRT2In)*funCF(v,SQRT2In)*(getCos(j,v,subY)*getCos(k,u,subX)));
						  
   
   /* 画像の縦横を8の倍数に正規化する(切り捨てか白を埋め込む)。*/
   
   nx = in->sx;
   ny = in->sy;
   
   Image *dummyOut = new Image(nx,ny);
   
   /* 初期化 */
   for(i=0;i<ny;i++)
     for(j=0;j<nx;j++){
       dummyOut->img[i][j] = 0.0;
    }
   
   
  
  
   /* mainの subX by subYブロックのloop */
   for(i=0;i<ny;i+=subY)
     for(j=0;j<nx;j+=subX){
       
       /* 画像をsubXxsubYの部分画像に分割する。*/
       for(l=0;l<subY;l++)
	 for(m=0;m<subX;m++)
	   subIn->img[l][m] = in->img[(i+l)][(j+m)];
      
       
       /*  部分画像１つに対してInverse DCTを行う。*/
       InverseDCTSub(subIn,subOut,weight1,TwoSQRTSubXSubYIn);
       
       for(l=0;l<subY;l++)
	for(m=0;m<subX;m++){
	  /* ↓DCTされた部分画像の周波数を保存用の配列に統合する。*/
	  dummyOut->img[(i+l)][(j+m)] = subOut->img[l][m];
	}
     }
   /* 出力の配列に代入 */
   for(i=0;i<out->sy;i++)
     for(j=0;j<out->sx;j++){
      
       out->img[i][j] = dummyOut->img[i][j];
      
     }
   
   delete weight1;
   
   delete subIn;
   delete subOut;
  
   delete dummyOut;
   
   
}



#undef PI
#undef funC
